 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef AGMAP_H
#define AGMAP_H

#include <map>
#include <vector>
#include <iostream>

using namespace std;
/**
 * This is a C++ port of Owen's Java AgMap class, here's the description
 * that Owen gives:
 * 
 *  aggregation map  OR  attribute grouping map. 
 *  For each attribute values in a rollup
 *  (of a possibly diced cube) it shows the base-cube attribute values
 *  corresponding to this derived attribute value.
 *
 *  Efficient calculation will probably "invert" this so that we
 *  build the derived cube higgledy-piggledy in one structured
 *  sweep through the base cube.
 *
 *  Unlike the Java code, this assumes the weighted case.
 *  
 *
 */
 

 
class InvertedAgMap {
        public: 
          
          // By default, we have the identity map. All ancestors are set to 
          //the identity (getNumberOfAncestors(12)= 12).
          InvertedAgMap() : mMap(0), mWeight(0), mNumberOfAttributeValues(0) {}
          
          InvertedAgMap(int NumberOfAttributeValues) : mMap(0), mWeight(0), 
            mNumberOfAttributeValues(NumberOfAttributeValues)
	    {}
          
          inline void setNumberOfAttributeValues(int NumberOfAttributeValues) {
                  mNumberOfAttributeValues = NumberOfAttributeValues;
          }
          	  
          inline void setAttributeValueAncestors(int attributevalue,
						 int  ancestors[], 
						 double ancestorWeights[],
						 int NumberOfAncestors){        
               vector<int> ancestorsvec(NumberOfAncestors); 
	       vector<double> ancestors_wvec(NumberOfAncestors);

               for(int k = 0; k < NumberOfAncestors; ++k) {
                ancestorsvec[k] = ancestors[k];
	        ancestors_wvec[k] = ancestorWeights[k];
	       }
               setAttributeValueAncestors(attributevalue,ancestorsvec, ancestors_wvec); 
          }
          
          inline void setAttributeValueAncestors(int attributevalue, 
						 const vector<int>& ancestors,
						 const vector<double>& ancestorsW){
            if(mMap.size() == 0) {
                //if(mNumberOfAttributeValues == 0) cerr<<" you should have called setNumberOfAttributeValues " <<endl;
                mMap.resize(mNumberOfAttributeValues);
		mWeight.resize(mNumberOfAttributeValues);
             }
             mMap[attributevalue] = ancestors; 
	     mWeight[attributevalue] = ancestorsW;
           }

	  
          
          inline bool isIdentity() const {return mMap.size() == 0;}

          inline vector<int> getAttributeAncestors(int attributevalue) const{
               if( isIdentity() ) cerr << "what are you doing???" << endl;
               return mMap[attributevalue];
           }
          
          inline uint getNumberOfAncestors(int attributevalue) const 
          {return mMap[attributevalue].size();}
          
          // these available for performance reasons
          inline vector<vector<int> >& getMap() {return mMap;}
	  inline vector<vector<double> >&getWeight() { return mWeight;}
          
          inline int getAttributeAncestor(int attributevalue, int number) const {
          //assert((attributevalue> 0) && (attributevalue < mMap.size()));
          return mMap[attributevalue][number];}

	  inline double getAttributeAncestorWeight(int attributevalue, int number) const {
	  return mWeight[attributevalue][number];}

          
        private:
          vector<vector<int> > mMap;
	  vector<vector<double> > mWeight;
          int mNumberOfAttributeValues;
};



 


#endif 
