// Lemur OLAP library (c) 2003 National Research Council of Canada by Daniel Lemire, and Owen Kaser
 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "fileutil.h"


//returns the size of an existing file in bytes
// WARNING: 0 if there is an error or if file doesn't exist
uint64 FileUtil::getFileSize(const char* fileName)
{
  if(fileExists	(fileName)) {
    struct stat my_stat;
    //assumes file exists (probably not the best strategy)// now ok?
    stat(fileName, &my_stat);
    if ((my_stat.st_size) > 0)
      return my_stat.st_size;
    else return 0;
  } else {
          return 0 ;// that's evil, but user probably doesn't use this method to check whether the file exists or not 
  }
}


// create all paths
int FileUtil::mkdir_parents(const char *path, mode_t mode) {
        int errnumero = mkdir(path,mode);
        if(vverbose) cout << " Trying to create " << path << endl;
        if ( errnumero < 0 && errnumero != EEXIST) {
                if(vverbose) cout << " failed... "<< errnumero<<  endl;
                if ((errnumero > 0 ) && (errnumero != ENOENT))  {// the path contains a file, no hope of solving this
                        cerr << "The path "<< path <<" could not be created, errnumero "<< errnumero<< ". No hope of creating it. " << endl;
                        return -1;
                }
                char * shorter_path = strdup(path);
                char * p = strrchr(shorter_path, '/');// find the last occurance of '/'
                if(p == 0 || p == path) { 
                  //cerr << " I won't be able to create " << shorter_path << endl;
                  free(shorter_path); 
                  return -1;
                }
                p[0] = 0;// set blank
                int returnvalue = mkdir_parents(shorter_path,mode);
                free(shorter_path);
                if(returnvalue == 0) return mkdir_parents(path,mode);
        }
        return 0;
}

// create all paths, but assume the path is a file
int FileUtil::mkdir_parentsonly(const char *path, mode_t mode) {
        //cout<<" called with "<<path<<endl;
        char * shorter_path = strdup(path);
        char * p = strrchr(shorter_path, '/');// find the last occurance of '/'
        if(p == 0 || p == path) { 
                  free(shorter_path); 
                  return 0;// nothing to do
        }
        p[0] = 0;
        //cout<<" making dir "<<shorter_path<<endl;
        int errnumero = mkdir(shorter_path,mode);
        if ( errnumero < 0 && errnumero != EEXIST) {
          if ((errnumero > 0 ) && (errnumero != ENOENT))  {// the path contains a file, no hope of solving this
                        cerr << "The path "<< path <<" could not be created, errnumero "<< errnumero<< ". No hope of creating it. " << endl;
                        free(shorter_path);
                        return -1;
          }
          int returnvalue = mkdir_parentsonly(shorter_path,mode);
          if(returnvalue == 0)
                  returnvalue = mkdir(shorter_path,mode);
          free(shorter_path);
          return returnvalue;
        } else {
                  free(shorter_path);
                  return 0;//success!
        }
}

//simple function to determine if a file exists
bool FileUtil::fileExists(const char* fileName)
{
   struct stat my_stat;
   return (stat(fileName, &my_stat) == 0);
}


time_t FileUtil::lastAccess(const char* fileName)
{
   struct stat my_stat;
   if (stat(fileName, &my_stat) == 0) {
           return my_stat.st_atime;
   }
   return -1;
}

time_t FileUtil::lastModification(const char* fileName)
{
   struct stat my_stat;
   if (stat(fileName, &my_stat) == 0) {
           return my_stat.st_mtime;
   }
   return -1;
}

void FileUtil::copy(const char * file1, const char * file2) {
  if(vverbose) cout << " orig. file size = " << FileUtil::getFileSize(file1) << endl;
  ofstream out(file2,std::ios::binary | std::ios::trunc );
  ifstream in(file1, std::ios::binary );
  const uint buffer_size = 256;
  char buffer[buffer_size];
  in.read((char*) & buffer, buffer_size);
  uint howmuchread = in.gcount();
  if(vverbose) cout << "howmuchread = " << howmuchread << endl;
  while( howmuchread  > 0 ) {
    out.write((const char*) &buffer,howmuchread);
    in.read((char*) & buffer, buffer_size);
    howmuchread = in.gcount();
    if(vverbose) cout << "howmuchread = " << howmuchread << endl;
  }
  if(vverbose) cout << "in.tellg() = " << in.tellg() << endl;
  assert(out.good());
  out.close();
  in.close();
  // no error handling  
}

// I am afraid both unlink and remove might be unixisms?
int FileUtil::unlink(const char * file) {
  //        return ::unlink(file);
        return ::remove(file);
}


void FileUtil::fillFile(const char * file, uint size) {
  ofstream out(file,std::ios::binary | std::ios::trunc );
  const uint buffer_size = 256;
  char buffer[buffer_size];
  memset(&buffer,0,buffer_size);
  uint64 howmuchwritten = 0;
  while(howmuchwritten < size) {
    uint tobewritten = MathUtil::min(buffer_size,size-howmuchwritten);
    out.write((const char *) & buffer, tobewritten);
    howmuchwritten += tobewritten;
    assert((uint64)out.tellp() == howmuchwritten);
  }
  assert(out.good());
  out.close();
}
