// Lemur OLAP library (c) 2003 National Research Council of Canada by Daniel Lemire, and Owen Kaser
 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#ifndef FILEUTIL_H
#define FILEUTIL_H

//#include <asm/errno.h>
#include <sys/stat.h> 
#include <sys/types.h>
#include <errno.h>
#include <malloc.h>
#include <cstring>
#include <iostream>
#include <fstream>
#include <cassert>
#include <cmath>
 

using namespace std;
typedef signed long long int64;
typedef unsigned long long uint64;
typedef unsigned int uint;


enum{ verbose = false, 
   vverbose = false};// vverbose is very verbose
   
class MathUtil {
  public:
    // seems like that's not defined in C++ by default... what a screwy language
    static inline int min (int x, int y ) { return x < y ? x : y;}
    // why on earth isn't that in the C++ api...
    static inline int max (int x, int y ) { return x > y ? x : y;}
};

class FileUtil {// always better to not put things in global namespace...
  public:
    static bool fileExists(const char* fileName);
    static int unlink(const char * file);//return 0 if ok
    static time_t lastAccess(const char* fileName);
    static time_t lastModification(const char* fileName);
    // returns 0 if ok
    /**
    * make the directory and all parents
    */
    static int mkdir_parents(const char *path, mode_t mode=0777);// 0777 is a potential security risk
    /**
    * only make the directories parent to this path
    */
    static int mkdir_parentsonly(const char *path, mode_t mode=0777);// 0777 is a potential security risk    
    static uint64 getFileSize(const char* fileName);
    // copy one file over the other one
    static void copy(const char * file1, const char *  file2);
    // open the file and fill it with NULL until it reaches the desired size
    static void fillFile(const char * file, uint size);


};


#endif
