#include "holapcube.h"


void test() {
  vector<uint> shape(3,4);
  vector<uint> chunkshape(3,2);
  //cout << " creating object "<< endl;
  HolapCube<int> hc (shape,chunkshape,"my_holaptest_directory");
  //cout << " query "<< endl;
  vector<uint> query(3,0);
  hc.put(query,3);
  if( hc.get(query) != 3) cout << " !!! error 3" << endl;
  //cout << " 3 == " << hc.get(query) << endl;
  //cout << " query "<< endl;
  vector<uint> query2(3,3);
  hc.put(query2,7);
  if( hc.get(query2) != 7) cout << " !!! error 7" << endl;
}

void fillingTest() {
  vector<uint> shape(3,4);
  vector<uint> chunkshape(3,2);
  //cout << " creating object "<< endl;
  HolapCube<int> hc (shape,chunkshape,"my_holaptest_directory");
  vector<uint> query(3,0);
  for(uint k = 0; k < 4 ; ++k) 
    for(uint l = 0; l < 4 ; ++l) 
      for(uint m = 0; m < 4 ; ++m) {
        query[0] = k; query[1] = l; query[2] = m;
        hc.put(query,k+l+m);
        int stored = hc.get(query);
        int stored2 = hc.get(query);
        if( stored != stored2) cout << "!!! not always returning the same value!" << endl;
        if( stored - (  k+l+m ) != 0) { 
          cout << "!!!! error at " << k << " " << l << " " << m <<  " should have stored " << k+l+m 
            << " but found " << stored << endl;
          cout << "I'm trying again... " << endl;
          hc.put(query,k+l+m);
          stored = hc.get(query);
          if( stored - (k+l+m) != 0) {
             cout<< "!!! Didn't fix the problem! " << endl;
          } else 
            cout << "The problem is solved by doing two puts" << endl;
        }
      }
  for(uint k = 0; k < 4 ; ++k) 
    for(uint l = 0; l < 4 ; ++l) 
      for(uint m = 0; m < 4 ; ++m) {
        query[0] = k; query[1] = l; query[2] = m;
        int stored = hc.get(query);
        if( stored - (  k+l+m ) != 0) 
          cout << "!!!! error while double checking at " << k << " " << l << " " << m << endl;
      }


 }

void iteratorTest() {
  vector<uint> shape(3,4);
  vector<uint> chunkshape(3,2);
  //cout << " creating object "<< endl;
  HolapCube<int> hc (shape,chunkshape,"my_holaptest_directory");
  CubeIterator<int> ci = hc.iterator(vector<uint>(3,1), vector<uint>(3,4));
  while(ci.next()) {
    //std::cout << " found non zero value = " <<  ci.value() << std::endl;
    int stored = hc.get(ci.key());
    if( stored == 0) {
      cout << " !!! iterator points at a zero value!" << endl;
    }
    if(stored != ci.value()) {
      cout << " !!! iterator value is wrong!" << endl;
    }
    hc.put(ci.key(),0); // clears the value
  } 
  CubeIterator<int> ci2 = hc.iterator(vector<uint>(3,1), vector<uint>(3,4)) ;
  if(ci2.next()) cout << " !!! iterator didn't detect all values?" << endl;
}

void iteratorBlank() {
  vector<uint> shape(3,4);
  vector<uint> chunkshape(3,2);
  //cout << " creating object "<< endl;
  HolapCube<int> hc (shape,chunkshape,"my_holaptest_directory");
  CubeIterator<int> ci = hc.iterator(vector<uint>(3,0), vector<uint>(3,4));
  while(ci.next()) { 
    hc.put(ci.key(),0); // clears the value
  } 
  for(uint k = 0; k < 4 ; ++k) 
    for(uint l = 0; l < 4 ; ++l) 
      for(uint m = 0; m < 4 ; ++m) {
        vector <uint> query(3);
        query[0] = k; query[1] = l; query[2] = m;
        int stored = hc.get(query);
        if(stored != 0) cout << "!!! blanking failed!" << endl;
      }
}


int main(char **) {
  cout << "starting holapcube testing"<<endl;
  test();
  cout << "done normal test, now startin filling test" << endl;
  fillingTest();
  cout << "done filling test, doing iterator test" << endl;
  iteratorTest();
  cout << "done iterator test, doing iterator blanking" << endl;
  iteratorBlank();
  cout << "repeating holapcube testing"<<endl;
  test();
  cout << "done normal test, now startin filling test" << endl;
  fillingTest();
  cout << "done filling test, doing iterator test" << endl;
  iteratorTest();
  cout << "done iterator test, doing iterator blanking" << endl;
  iteratorBlank();
  cout << "if you made it that far, then I hope there is no bug."<< endl;
  return 0;
}
