#include "bigarray.h"
#include "internalbigarray.h"
#include "btreecube.h"
#include "cdbcube.h"
#include <time.h>
#include <cmath>
#include <stdlib.h>

/**
* This file serves to test the speed of slicing.
* Daniel Lemire, August 12th, 2005
**/
int main(char **) {
        char * bigfile = "bigchunk.bin";
        clock_t start,finish;
        double seconds;
        int d = 3;
        int n = 10000;// cube is n*n*n
        int records = 100000;//1 million!
        cout << "Building a cube with "<<records<<" records in it, and "<< d <<" dimensions."<<endl;
        ::unlink(bigfile);
        start = clock();
        BTreeCube<int> bt(d,bigfile);
        cout << "Filling the cube with random content..." << endl;
        // fill the file with random contant
        vector<int> pos(d);
        srand(start);
        for (int k = 0; k < records; ++k) {
                int times = 0;
                //cout<< "entering record " << k << endl;
                while(true) {
                        for (int dim = 0; dim < d;++ dim)
                               pos[dim] = (int) floor((double)rand()*n/RAND_MAX);
                        //cout << "got the location " << endl;
                        if(bt.get(pos) == 0) {
                               bt.put(pos,1);
                               break;
                        }
                        ++ times;
                        if(times > 2) {
                                cout << "cube too dense, aborting"<<endl;
                                break;
                        }
                }
                if(k %10000 == 0)  cout << k << " records written... ("<< 100.0*k/records <<"%) "<< endl;
        }
        finish = clock();
        seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "Cube completed in "<< seconds << endl;
        cout << "Number of records : "<< bt.getNumberOfRecords() << endl;
        int numberoftests = 10;        
        for (int dim = 0; dim < d; ++dim) {
                cout << endl << " =============================" << endl;
                cout << " Trying to slice on dimension " << dim << endl;
                float totaltime = 0.0;
                for(int times = 0; times < numberoftests ; ++times) {
                        int val = (int) floor((double)rand()*n/RAND_MAX);
                        ::unlink("slicecube.bin");
                        start = clock();
                        MultidimensionalArray<int>* sliced= bt.slice(dim, val,"slicecube.bin");
                        //int slicedrecords  = sliced->getNumberOfRecords();
                        delete sliced;
                        finish = clock();
                        seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
                        totaltime += seconds;
                        //cout << "time = "<< seconds<< " s for "<<slicedrecords<< " records" <<endl; 
                }
                cout << "average time is "<< totaltime/numberoftests<< " s"<< endl;
        }

}
