#include "bigarray.h"
#include "internalbigarray.h"
#include "btreecube.h"
//#ifdef USE_CDB
#include "cdbcube.h"
//#endif
#include <time.h>

/**
* this test whether I can create a subdirectory automatically!
*/
template <class cubetype>
void testcreatedirs() {
        vector<int> bigshape(3);
        bigshape[0]=50;bigshape[1]=50;bigshape[2]=50;
        if(FileUtil::fileExists("tempdirectory/anotherdirectory/file1.bin")) {
           ::unlink("tempdirectory/anotherdirectory/file1.bin");
           if(::rmdir("tempdirectory/anotherdirectory")!=0)  {
             cerr<<endl<<endl<<endl<<"could not delete tempdirectory/anotherdirectory, please investigate!!!!*******"<<endl<<endl<<endl<<endl<<endl;
             return;
           }
           if(::rmdir("tempdirectory")!=0) {
             cerr<<endl<<endl<<endl<<endl<<"could not delete tempdirectory, please investigate!!!!*******"<<endl<<endl<<endl<<endl<<endl;
             return;
           }
        }
        cubetype bt(bigshape, "tempdirectory/anotherdirectory/file1.bin");
        bt.close();
        ::unlink("tempdirectory/anotherdirectory/file1.bin");
        ::rmdir("tempdirectory/anotherdirectory");
        ::rmdir("tempdirectory");
}

template <class cubetype>
void longtest() {
        vector<int> bigshape(3);
        bigshape[0]=500;bigshape[1]=500;bigshape[2]=50;
        cubetype bt(bigshape);
        cout<< endl << endl<< "==longtest== with "<< bt.cubeType()<<endl;
        cout << " filling with zeroes... "<<endl;
        bt.fillWithZeroes();  
        cout << " setting..." << endl;
        {
        clock_t start = clock();
        int total = 0;
        vector<int> coordinates(3);
        for(int x = 0; x < 50; ++x) {
                for(int y = 0; y < 5; ++y) {
                        for(int z= 0; z< 50; ++z) {
                                coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                bt.put(coordinates,x+y+z);
                                ++total;
                        }
                }
        }
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SET test took " << seconds << " for " << total << " sets"<< endl;
        }
        cout<<" getting..." <<endl;
        {
        clock_t start = clock();
        int total = 0;
        vector<int> coordinates(3);
        for(int x = 0; x < 50; ++x) {
                for(int y = 0; y < 5; ++y) {
                        for(int z= 0; z< 50; ++z) {
                                coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                if( bt.get(coordinates)!= x+y+z) {
                                        cerr <<endl<<endl<<endl<<endl<<"BUG!!!!" <<endl;
                                        cerr << " I was expecting : "<< x+y+z<< endl;
                                        cerr << " I got : "<< bt.get(coordinates)<< endl <<endl<<endl<<endl<<endl;
                                }
                                ++total;
                        }
                }
        }
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "GET test took " << seconds << " for " << total << " sets"<< endl;
        }
        {
        cout << "Running ICEBERG test..."<< endl;
        bool verbose = false;
        clock_t start = clock();
                try {
                        if(verbose) cout << "topn freq ..."<<endl;
                        int top = bt.topN_frequency(10);
                        if(verbose) cout<< "top 10:" << top << endl;
                        if(verbose) cout << "computing iceberg..."<<endl;
                        ::unlink("ice");
                        cubetype * bt2 = (cubetype*) bt.iceberg( top,  "ice");
                        if(bt2->getNumberOfRecords() != 10) {
                                cerr <<"**************************"<<endl;
                                cerr <<"**************************"<<endl;
                                cerr <<"BUG: expected "<<10 <<" records got "<< bt2->getNumberOfRecords()<<endl;
                        }
                        assert (bt2->deleteOnDisk());
                
                } catch (std::out_of_range e) {
                        if(verbose) cout << " no possible top 10" << endl;
                }
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "ICEBERG test took " << seconds << " seconds"<< endl;
        }
        {
                cout << "Running SLICE test..."<< endl;
                clock_t start = clock();
                bool verbose = false;
                for(int times = 0; times < 10; ++times) {
                        if(verbose) cout << "slicing...(first dim)"<<endl;
                        ::unlink("slice");
                        cubetype * bt3 = (cubetype*) bt.slice( 0,times,  "slice");
                        if(verbose) cout << "topn freq over slice..."<<endl;
                        try {
                                if(verbose) cout<< "top 10:" << bt3->topN_frequency(10)<< endl;
                        } catch (std::out_of_range f ) {
                                //cout<< "no iceberg with 10 or less elements."<< endl;
                        }
                        assert (bt3->deleteOnDisk());
                        if(verbose) cout << "slicing... (second dim)"<<endl;
                        ::unlink("slice");
                        cubetype * bt4 = (cubetype*) bt.slice( 1,times,  "slice");
                        if(verbose) cout << "topn freq over slice..."<<endl;
                        try {
                                if(verbose) cout<< "top 10:" << bt4->topN_frequency(10)<< endl;
                        } catch (std::out_of_range g) {
                                if(verbose) cout<< "no iceberg with 10 or less elements."<< endl;
                        }
                        assert (bt4->deleteOnDisk());
                }
                clock_t finish = clock();
                double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
                cout << "SLICE test took " << seconds << " seconds"<< endl;
       }
}

template <class cubetype>
void testglue() {
        vector<int> bigshape(3);
        bigshape[0]=500;bigshape[1]=500;bigshape[2]=50;
        cubetype bt(bigshape);
        cout<< endl << endl<< "==gluetest== with "<< bt.cubeType()<<endl;
        cout << " filling with zeroes... "<<endl;
        bt.fillWithZeroes();  
        const int zmax = 10;
        cout << " setting..." << endl;
        {
        clock_t start = clock();
        int total = 0;
        vector<int> coordinates(3);
        for(int x = 0; x < 50; ++x) {
                for(int y = 0; y < 5; ++y) {
                        for(int z= 0; z< zmax; ++z) {
                                coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                bt.put(coordinates,x+y+z);
                                ++total;
                        }
                }
        }
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SET test took " << seconds << " for " << total << " sets"<< endl;
        }
        {
        cout << "Running SLICE/GLUE test..."<< endl;
        clock_t start = clock();
        CubeSet<int> btslice;
        for(int z= 0; z < zmax; ++z) {
                stringstream strs;
                strs << "slice"<<z;
                ::unlink(strs.str().c_str());
                btslice.add(*bt.slice(2,z, strs.str()));
        }
        cout << "SLICING completed... glueing..."<< endl;
        ::unlink("superglue");
        CubeSet<int> others(btslice);
        others.remove(0); 
        cubetype * master =(cubetype *) btslice.get(0).glue(others,"superglue");
        for(int z= 0; z < zmax; ++z) delete &btslice.get(z);
        delete master;
        cout << "GLUE and slice completed..."<< endl;
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SLICE and GLUE test took " << seconds << " seconds"<< endl;
                cout<<" checking..." <<endl;
                {
                        vector<int> coordinates(3);
                        for(int x = 0; x < 50; ++x) {
                                for(int y = 0; y < 5; ++y) {
                                        for(int z= 0; z< zmax; ++z) {
                                                coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                                if( bt.get(coordinates)!= x+y+z) {
                                                        cerr <<endl<<endl<<endl<<endl<<"BUG!!!!" <<endl;
                                                        cerr << " I was expecting : "<< x+y+z<< endl;
                                                        cerr << " I got : "<< bt.get(coordinates)<< endl <<endl<<endl<<endl<<endl;
                                                }
                                        }
                                }
                        }
                }
                cout<<" checking...ok" <<endl;
         }

}

template <class cubetype>
void longtestarrays() {        
        vector<int> bigshape(3);
        bigshape[0]=500;bigshape[1]=500;bigshape[2]=50;
        cubetype bt(bigshape);
        cout<< endl << endl<< "==longtestarrays== with "<< bt.cubeType()<<endl;
        cout << " filling with zeroes... "<<endl;
        bt.fillWithZeroes();  
        int xlength = 150;
        int ylength = 10;
        cout<<"[arrays] setting..." <<endl;
        {
        clock_t start = clock();
        int total = 0;
        int coordinates[3];
        for(int x = 0; x < xlength; ++x) {
                for(int y = 0; y < ylength; ++y) {
                        for(int z= 0; z< 50; ++z) {
                                coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                bt.put(coordinates,x+y+z);
                                ++total;
                        }
                }
        }
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SET test took " << seconds << " for " << total << " sets"<< endl;
        }
        cout<<"[arrays] getting..." <<endl;
        {
        clock_t start = clock();
        int total = 0;
        int coordinates[3];
        for(int x = 0; x < xlength; ++x) {
                for(int y = 0; y < ylength; ++y) {
                        for(int z= 0; z< 50; ++z) {
                                coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                if( bt.get(coordinates)!= x+y+z) {
                                        cerr <<endl<<endl<<endl<<endl <<"BUG!!!!" <<endl;
                                        cerr << " I was expecting : "<< x+y+z<< endl;
                                        cerr << " I got : "<< bt.get(coordinates)<< endl <<endl<<endl<<endl<<endl;
                                }
                                ++total;
                        }
                }
        }        
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "GET test took " << seconds << " for " << total << " sets"<< endl;
        }
        {
                cout << "copying the cube to a CDB cube..." <<endl;
                CDBCube<int> cdb(bt);
                cout << "ok, now testing get test!"<<endl; 
                
                clock_t start = clock();
                int total = 0;
                int coordinates[3];
                for(int x = 0; x < xlength; ++x) {
                        for(int y = 0; y < ylength; ++y) {
                                for(int z= 0; z< 50; ++z) {
                                        coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                        if( cdb.get(coordinates)!= x+y+z) {
                                                cerr <<endl<<endl<<endl<<endl <<"BUG!!!!" <<endl;
                                                cerr << " I was expecting : "<< x+y+z<< endl;
                                                cerr << " I got : "<< cdb.get(coordinates)<< endl <<endl<<endl<<endl<<endl;
                                        }
                                        ++total;
                                }
                        }
                }        
                clock_t finish = clock();
                double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
                cout << "GET test over CDB took " << seconds << " for " << total << " sets"<< endl;
        cout << "Look how fast CDB cubes are!!! "<<endl;
        }

}





template <class cubetype>
void longtestiterators() {        
        vector<int> bigshape(3);
        bigshape[0]=500;bigshape[1]=500;bigshape[2]=50;
        cubetype bt(bigshape);
        cout<< endl << endl<< "==longtestiterators== with "<< bt.cubeType()<<endl;
        cout << " filling with zeroes... "<<endl;
        bt.fillWithZeroes();
        int xlength = 50;
        int ylength = 5;
        int zlength = 50;
        cout<<"[it] setting..." <<endl;
        {
        clock_t start = clock();
        int total = 0;
        int coordinates[3];
        for(int x = 0; x < xlength; ++x) {
                for(int y = 0; y < ylength; ++y) {
                        for(int z= 0; z< zlength; ++z) {
                                coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                bt.put(coordinates,x+y+z+1);
                                ++total;
                        }
                }
        }
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SET test took " << seconds << " for " << total << " sets"<< endl;
        }
        cout<<"[it] iteration..." <<endl;
        {
        clock_t start = clock();
        int total = 0;
        MultidimensionalIterator<int> *it = bt.iterator();
        while(it->next()) {
          vector<int> coordinates = it->key();
          if(it->value() != coordinates[0]+coordinates[1]+coordinates[2]+1) {
                                        cerr  <<endl<<endl<<endl<<endl<<"BUG!!!!" <<endl <<endl<<endl<<endl<<endl;
                                        cerr <<" at line "<<__LINE__<<endl;
          }
                                        
          ++total;
        }
        //cout << " got "<< total << "data" << " xlength = "<< xlength << " ylength = "<<ylength << endl;
        assert(xlength*ylength*zlength == total);  
        delete it;
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "ITER test took " << seconds << " for " << total << " data"<< endl;
        }

}



template <class cubetype>
void longtestsliceiterators() {        
        vector<int> bigshape(3);
        bigshape[0]=500;bigshape[1]=500;bigshape[2]=50;
        cubetype bt(bigshape);
        cout<< endl << endl<< "==longtestsliceiterators== with "<< bt.cubeType()<<endl;
        cout << " filling with zeroes... "<<endl;
        bt.fillWithZeroes();  
        int xlength = 100;
        int ylength = 100;
        cout<<"[it] setting..." <<endl;
        {
        clock_t start = clock();
        int total = 0;
        int coordinates[3];
        for(int x = 0; x < xlength; ++x) {
                for(int y = 0; y < ylength; ++y) {
                        for(int z= 0; z< bigshape[2]; ++z) {
                                coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                bt.put(coordinates,x+y+z+1);
                                ++total;
                        }
                }
        }
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SET test took " << seconds << " for " << total << " sets"<< endl;
        }
        //
        //
        cout<<"[it] slice iteration..." <<endl;
        {
        clock_t start = clock();
        int total = 0;
        int zvalue = 10;
        MultidimensionalIterator<int> *it = bt.sliceIterator(2,zvalue);
        while(it->next()) {
          vector<int> coordinates = it->key();
          assert(coordinates[2] == zvalue);
          if(it->value() != coordinates[0]+coordinates[1]+zvalue+1) {
                                        cerr <<endl<<endl<<endl<<endl<<"BUG!!!!" <<endl <<endl<<endl<<endl<<endl;
                                        cerr <<" at line "<<__LINE__<<endl;
                                        cerr << " in cube, at "<<coordinates[0]<< ", "<<coordinates[1]<< ", "<<coordinates[2]<< " I got "<< it->value()<< ", I was expecting "<< coordinates[0]+coordinates[1]+zvalue+1<< endl;
          }
          ++total;
        }

        if( total != ylength * xlength) {
                cerr <<endl<<endl<<endl<<endl<<"BUG!!!!" <<endl <<endl<<endl<<endl<<endl;
                cerr <<" at line "<<__LINE__<<endl;
                cout << "total = "<<total<<endl;
                cout << "ylength = "<<ylength<< " xlength = "<< xlength<<endl;
        }
        delete it;
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SLICE ITER test took " << seconds << " for " << total << " sets"<< endl;
        }
        //
        //
        cout<<"[it] slice iteration using range query..." <<endl;
        {
        clock_t start = clock();
        int total = 0;
        int zvalue = 10;
        vector<int> s(3,0);  s[2]= zvalue;
        vector<int> e(bigshape);  e[2] = 1 +  zvalue;
        MultidimensionalIterator<int> *it = bt.iterator(s,e);
        while(it->next()) {
          vector<int> coordinates = it->key();
          assert(coordinates[2] == zvalue);
          if(it->value() != coordinates[0]+coordinates[1]+zvalue+1) {
                                        cerr <<endl<<endl<<endl<<endl<<"BUG!!!!" <<endl <<endl<<endl<<endl<<endl;
                                        cerr <<" at line "<<__LINE__<<endl;
                                        cerr << " in cube, at "<<coordinates[0]<< ", "<<coordinates[1]<< ", "<<coordinates[2]<< " I got "<< it->value()<< ", I was expecting "<< coordinates[0]+coordinates[1]+zvalue+1<< endl;
          }
          ++total;
        }       
        if( total != ylength * xlength) {
                cerr <<endl<<endl<<endl<<endl<<"BUG!!!!" <<endl <<endl<<endl<<endl<<endl;
                cerr <<" at line "<<__LINE__<<endl;
                cout << "total = "<<total<<endl;
                cout << "ylength = "<<ylength<< " xlength = "<< xlength<<endl;
        }
        delete it;
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "EMULATED SLICE ITER test took " << seconds << " for " << total << " sets"<< endl;
        }
        cout << "Note: sometimes I get that SLICE ITER is much slower than EMULATED SLICE ITER and the explanation has to do with the memory file mapping!!! To test the theory, I'm running the test a second time." <<endl;
        cout<<"[it] slice iteration... (test 2)" <<endl;
        {
        clock_t start = clock();
        int total = 0;
        int zvalue = 10;
        MultidimensionalIterator<int> *it = bt.sliceIterator(2,zvalue);
        while(it->next()) {
          vector<int> coordinates = it->key();
          assert(coordinates[2] == zvalue);
          if(it->value() != coordinates[0]+coordinates[1]+zvalue+1) {
                                        cerr <<endl<<endl<<endl<<endl<<"BUG!!!!" <<endl <<endl<<endl<<endl<<endl;
                                        cerr <<" at line "<<__LINE__<<endl;
                                        cerr << " in cube, at "<<coordinates[0]<< ", "<<coordinates[1]<< ", "<<coordinates[2]<< " I got "<< it->value()<< ", I was expecting "<< coordinates[0]+coordinates[1]+zvalue+1<< endl;
          }
          ++total;
        }

        if( total != ylength * xlength) {
                cerr <<endl<<endl<<endl<<endl<<"BUG!!!!" <<endl <<endl<<endl<<endl<<endl;
                cerr <<" at line "<<__LINE__<<endl;
                cout << "total = "<<total<<endl;
                cout << "ylength = "<<ylength<< " xlength = "<< xlength<<endl;
        }
        delete it;
        clock_t finish = clock();
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SLICE ITER test took " << seconds << " for " << total << " sets"<< endl;
        }
        cout << "Let's see how fast we can slice and create a cube!!!"<<endl;
        {
        ::unlink("garbageslicing.bin");
        clock_t start = clock();
        int zvalue = 10;
        //MultidimensionalArray<int> *it =
        for(int k = 0;k < 5;++k)
	  delete(bt.slice(2,zvalue,"garbageslicing.bin"));
        clock_t finish = clock();
        ::unlink("garbageslicing.bin");
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SLICE ITER + cube creation test took " << seconds << endl;
        }   
        cout << "Let's see how fast we can slice and create a CDB cube!!!"<<endl;
        {
        ::unlink("garbageslicing.bin");
        clock_t start = clock();
        int zvalue = 10;
        //MultidimensionalArray<int> *it =
        for(int k = 0;k < 5;++k)
	  delete(bt.slice(2,zvalue,"garbageslicing.bin",MultidimensionalArray<int>::CUBETYPE_CDBCUBE));
        clock_t finish = clock();
        ::unlink("garbageslicing.bin");
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "SLICE ITER + cube CDB creation test took " << seconds << endl;
        }
        cout << "Let's see how fast we can rollup/slice/iceberg and create a cube!!!"<<endl;
        {
        ::unlink("garbagerollup.bin");
        ::unlink("garbagerollupslicing.bin");
        ::unlink("garbagerollupslicingice.bin");
        clock_t start = clock();
        int zvalue = 10;
        //MultidimensionalArray<int> *it =
        MultidimensionalArray<int> * rollup = bt.full_rollup(vector<int>(1,1),"garbagerollup.bin");
        clock_t top1 = clock();
        MultidimensionalArray<int> * rollupslice = rollup->slice(1,zvalue,"garbagerollupslicing.bin");
        clock_t top2 = clock();
        delete rollup;
        int threshold = rollupslice->topN_frequency(10);
        clock_t top3 = clock();
        MultidimensionalArray<int> * rollupsliceice = rollupslice->iceberg(threshold,"garbagerollupslicingice.bin");  
        delete rollupslice;
        delete rollupsliceice;
        clock_t finish = clock();
        ::unlink("garbagerollup.bin");
        ::unlink("garbagerollupslicing.bin");
        ::unlink("garbagerollupslicingice.bin");
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "ROLLUP/SLICE/TOPN/ICE + cube creation test took " << seconds << endl;
        cout << "ROLLUP= " << (double)(top1 - start) / CLOCKS_PER_SEC << endl;
        cout << "SLICE= " << (double)(top2 - top1) / CLOCKS_PER_SEC << endl;
        cout << "TOPN= " << (double)(top3 - top2) / CLOCKS_PER_SEC << endl;
        cout << "ICEBERG= " << (double)(finish - top3) / CLOCKS_PER_SEC << endl;
        }   
        cout << "Let's see how fast we can rollup/slice/iceberg and create a CDB cube!!!"<<endl;
        {
        ::unlink("garbagerollup.bin");
        ::unlink("garbagerollupslicing.bin");
        ::unlink("garbagerollupslicingice.bin");
        clock_t start = clock();
        int zvalue = 10;
        //MultidimensionalArray<int> *it =
        MultidimensionalArray<int> * rollup = bt.full_rollup(vector<int>(1,1),"garbagerollup.bin",MultidimensionalArray<int>::CUBETYPE_BTREECUBE);//rollup doesn't support cdbcube!!!
        clock_t top1 = clock();
        MultidimensionalArray<int> * rollupslice = rollup->slice(1,zvalue,"garbagerollupslicing.bin",MultidimensionalArray<int>::CUBETYPE_CDBCUBE);
        clock_t top2 = clock();
        delete rollup;
        int threshold = rollupslice->topN_frequency(10);
        clock_t top3 = clock();
        MultidimensionalArray<int> * rollupsliceice = rollupslice->iceberg(threshold,"garbagerollupslicingice.bin",MultidimensionalArray<int>::CUBETYPE_CDBCUBE);  
        delete rollupslice;
        delete rollupsliceice;
        clock_t finish = clock();
        ::unlink("garbagerollup.bin");
        ::unlink("garbagerollupslicing.bin");
        ::unlink("garbagerollupslicingice.bin");
        double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
        cout << "ROLLUP/SLICE/TOPN/ICE + CDBcube creation test took " << seconds << endl;
        cout << "ROLLUP= " << (double)(top1 - start) / CLOCKS_PER_SEC << endl;
        cout << "SLICE= " << (double)(top2 - top1) / CLOCKS_PER_SEC << endl;
        cout << "TOPN= " << (double)(top3 - top2) / CLOCKS_PER_SEC << endl;
        cout << "ICEBERG= " << (double)(finish - top3) / CLOCKS_PER_SEC << endl;
        }   
        cout << "Now, I will try my hand at CDBCube and see how fast it can slice and iterate!"<<endl;
        {
                cout <<"creating and loading the constant cube from the existing cube (timing performance of both the original and the CDBCube)..." <<endl;
                ::unlink("CDBCubeTempSuperSpecial.bin");
                clock_t start = clock();
                CDBCube<int> cdb(bt, "CDBCubeTempSuperSpecial.bin");
                clock_t finish = clock();
                cout << "Loaded!"<<endl;
                double seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
                cout << "Cube loading took " << seconds <<endl;
                cout <<"Now checking its content..." <<endl;
                start = clock();
                int coordinates[3];
                for(int x = 0; x < xlength; ++x) {
                        for(int y = 0; y < ylength; ++y) {
                                for(int z= 0; z< bigshape[2]; ++z) {
                                        coordinates[0]=x;coordinates[1]=y;coordinates[2]=z;
                                        assert(cdb.get(coordinates) ==x+y+z+1);
                                }
                        }
                }
                finish = clock();
                seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
                cout << "CDB Cube 'getting' took " << seconds << endl;
                ::unlink("garbageslicing.bin");
                start = clock();
                int zvalue = 10;
                for(int k = 0;k < 5;++k)
	          delete(bt.slice(2,zvalue,"garbageslicing.bin",MultidimensionalArray<int>::CUBETYPE_CDBCUBE));
                finish = clock();
                ::unlink("garbageslicing.bin");
                seconds =  (double)(finish - start) / CLOCKS_PER_SEC;
                cout << "SLICE ITER + CDB cube creation over CDB  took " << seconds << endl;
                ::unlink("CDBCubeTempSuperSpecial.bin");

        }
        cout << "The story regarding CDBCubes is that they are much faster at queries but because keys are not sorted, the iterators are not as smart. Also, we don't have an iterator able to go seek a given location with a CDBCube. It is really raw force. "<<endl;
 
}

/* When C++ bombs on some special op invoked from Java
 * I haven't figured out how to debug.   This is the moral
 * equivalent of junit testing for the C++ ops.
 * -owen
 */

void specialOwenTests() {
  vector<int> query(3,0);
  vector<int> shape(3);
  shape[0]=3;shape[1]=4;shape[2]=2;
  BTreeCube<int> bt(shape.size());        
  cout << " filling with zeroes... "<<endl;
  bt.fillWithZeroes();
  cout << " 0 == " << bt.get(query) << endl;
  bt.put(query,10);   
  vector<int> at200(query); at200[0]=2;
  vector<int> at210(at200); at210[1]=1;
  vector<int> at220(at200); at220[1]=2;
  vector<int> at230(at200); at230[1]=3;
  vector<int> at221(at220); at221[2]=1;
  vector<int> at231(at230); at231[2]=1;
  cout << "starting bunch of puts..." <<endl;
  bt.put(at200, 20);
  bt.put(at210, 40);
  bt.put(at220, 80);
  bt.put(at230, 160);
  bt.put(at221, 320);
  bt.put(at231, 640);
  cout << "done bunch of puts..." <<endl;
  
  // leave dimension 0 alone, dim 1: put 0 and 1 together into new 0, put 1 and 2 and 3 into new 1.
  // dim 2: collapse
  
  // Constant initializers of ragged arrays isn't working for me so...
  int z[] = {0};  double zw[] = {1.0};
  int o[] = {1};  double ow[] = {1.0};
  int zo[] = {0,1}; double zow[] = {1.0,1.0};
  
  InvertedAgMap agmap[3] = {InvertedAgMap(3),InvertedAgMap(4),InvertedAgMap(2)}; // we have three identity maps
  //agmap[0] will remain the identity
  agmap[1].setAttributeValueAncestors(0, z, zw, 1);// last index is length of array
  agmap[1].setAttributeValueAncestors(1,zo, zow, 2);
  agmap[1].setAttributeValueAncestors(2,o,ow, 1);
  agmap[1].setAttributeValueAncestors(3,o,ow, 1);
  agmap[2].setAttributeValueAncestors(0,z,zw, 1);
  agmap[2].setAttributeValueAncestors(1,z, zw, 1);

  MultidimensionalArray<int> *bt1temp = bt.partial_rollup(agmap,string("/tmp/deleteme"));

  BTreeCube<int> *bt1 = dynamic_cast<BTreeCube<int>*>(bt1temp);
  cout << "10 == " << bt1->get(query) << endl;
  vector<int>at110(3,1);at110[2]=0;
  cout << "0 == " << bt1->get(at110) << endl;
  cout << "60 == " << bt1->get(at200) << endl;
  cout << "1240 == " << bt1->get(at210) << endl;

  vector<int>singleton0(1,0);  // kill dimension 0.

  MultidimensionalArray<int> *bt2 = bt1->full_rollup(singleton0,string("/tmp/deletemetoo"));

  vector<int>at00(2,0);
  cout << "70==" << bt2->get(at00) << endl;
  vector<int>at10(at00); at10[0]=1;
  cout << "1240 == " << bt2->get(at10) << endl;

#if 1
  // try resizing the thing to 10 x 10

  //int funnysize[] = {10,10};
  BTreeCube<int> bt3(*dynamic_cast<BTreeCube<int>*>(bt2));//, funnysize ,2, string("/tmp/deletemetrhee"));
  cout << "1240 == " << bt3.get(at10) << endl;
#endif


  MultidimensionalArray<int> *bt4 = bt.map_pointwise( Sum<int>(), bt, string("/tmp/deletemefoour"));
  cout << bt.get(query) << " is doubled to " << bt4->get(query) << endl;

  int shouldbe320 = bt.topN_frequency(2);
  cout << "second largest value in bt is " << shouldbe320 << " should be 320" << endl;
  assert (shouldbe320 == 320);
  
  cout << "calling slice on bt..." << endl;

  MultidimensionalArray<int> *bt5 = bt.slice(1,2,string("/tmp/deletemefive"));  // dim 1 value 2
  cout << "after slicing bt, result should have 80@2,0 and 320@2,1 " << endl;
  bt5->printData();

  int fooo[bt.getShape().size()];
  for (uint i = 0; i < bt.getShape().size(); ++i)
    fooo[i] = bt.getShape()[i];


  MultidimensionalArray<int> *bt6 = bt.map_1dvectors('c', bt, 4, 1, 1, 0, 0,  string("/tmp/deletemesix"));
  cout << "after doing scaled cosines, expect 100 at 2,0,2,0 and 2,1,2,1 "
          "and smaller values at 2,1,2,0 and [same] at 2,0,2,1 "<< endl;
  bt6->printData();


  // is all this bt aliasing mean?

  cout << " bt has count " << bt.getCountTemp();
  //const int ttt[] = {0,1000,2000};
  //MultidimensionalArray<int> * const tttt[] = {&bt,&bt,&bt,&bt};  // but I only use 2 of them!
  cout << " after four copies in an array, bt has count " << bt.getCountTemp() << endl;


  //  const BTreeCube<int> &foo = tttt[0];
  //  const BTreeCube<int> &bar = tttt[1];
  
  // cout << "foo=" << &foo << " bar=" << &bar << endl;
  // cout << "fooname=" << foo.getName() << " barname=" << bar.getName() << endl;
 

  //MultidimensionalArray<int> *bt7 = bt.collapse(tttt,2,ttt,1,string("/tmp/deletemeseven"));
  //cout << "after 3-way collapse we have " << endl;
  //bt7->printData();

  cout << " reading to call Daniel";
  // MultidimensionalArray<int> ttttt[] = {bt, bt};  // won't compile, type is virtual
  //BTreeCube<int> ttttt[] = {bt, bt};  // won't work, sizeof() differs from callee

  //bt7->daniel(ttttt);

}


int main(char **) {
        ::unlink("BTreeCubeTemp.bin");
  vector<int> shape(3);
  shape[0]=3;shape[1]=4;shape[2]=2;

  //
  //

  cout << endl<< "===starting internalbigarray testing==="<< endl<<endl;  
  {
  InternalBigArray<int> ba (shape);  
  ba.fillWithZeroes();
  vector<int> query(3,0);
  assert(0==ba.get(query));
  ba.put(query,1);
  assert(1==ba.get(query));}

#ifdef USECDB
  cout << endl<< "===starting cdb minimal testing==="<< endl<<endl;  
  {
  InternalBigArray<int> ba (shape);  
  ba.fillWithZeroes();
  vector<int> query(3,0);
  assert(0==ba.get(query));
  ba.put(query,1);
  assert(1==ba.get(query));
  CDBCube<int> cdb (ba);
  assert(1==cdb.get(query));
  }
#endif

  specialOwenTests();

  cout<< endl << "===starting btreecube serious testing==="<< endl<<endl;
  {
testglue<BTreeCube<int> >();
testcreatedirs<BTreeCube<int> >();
longtestarrays<BTreeCube<int> >();
longtest<BTreeCube<int> >();
longtestiterators<BTreeCube<int> >();
longtestsliceiterators<BTreeCube<int> >();
// we do it twice, just to be sure
longtestsliceiterators<BTreeCube<int> >();
  }


  cout << endl<< "===starting bigarray serious testing==="<<endl<<endl;
  {
testglue<BigArray<int> >();
testcreatedirs<BigArray<int> >();
longtestarrays<BigArray<int> >();
longtest<BigArray<int> >();
longtestiterators<BigArray<int> >();
longtestsliceiterators<BigArray<int> >();
//we do it twice, to be sure
longtestsliceiterators<BigArray<int> >();
  }
          
  
  
}
