 /**
 *  This program is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation (version 2). This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
 
#ifndef UBTreeCube_H
#define UBTreeCube_H



#include <string>
#include <fstream>
#include <cassert>
#include <stdexcept>
#include <vector>
#include <stdlib.h>
#include <stdio.h>
#include <sstream>
#include "fileutil.h"
#include "multidimensionalarray.h"
#include "bigarray.h"
#include <villa.h> // QDBM, see http://qdbm.sourceforge.net/, please download and install!, RPMs are available for RedHat-like systems
// Daniel: QDBM has a C++ API, but I found that it was somewhat limited (I complained
// to the author and maybe it is fixed now...)
// Also, we use villa because it has sorted keys, this means we can

using namespace std;



//template <class DataType>
//class CubeIterator;


/***
 * Consider a "sparse cube", that is a map from coordinate tuples to values
 * It is easy to store such a "sparse cube" in a B-tree using a DBM.
 * UB-Tree or "Universal B-Tree" map tuples to Z-indexes and use these indexes
 * as keys in the B-Tree. 
 * 
 * (This is currently non-functional code. It should be derived from btreecube.)
 *
 * See also btreecube.h.
 *
 * TODO: write an iterator over a range.
 *
 * author: Daniel Lemire, February 2005
 * 
 */
template<class DataType>
class UBTreeCube : public MultidimensionalArray {

 
  public:

    /**
     * Create a cube with chunks of a given shape. The data will be stored in 
     * a file (you must have write permissions to it, it will be created
     * if it doesn't exist). 
     */
    UBTreeCube(vector<uint> shape, string filename );
    
    virtual ~UBTreeCube();
    
   
    /**
     * In very rare cases, you might want to close the cube to 
     * release system resources. (Destructor will do it for you...)
     */
    void close();
   
    /**
     * If you ever happened to close the cube, you might need to open it.
     */
    void open();
    /**
    * erase all 
    */
    void fillWithZeroes();
    
    /**
     * Query the cube. If the cell is not allocated, returns 0.
     */
    DataType get( const vector<uint>& pos) const;

    /**
     * put a value at a given cell. If the value is 0, then 
     * it should delete the cell.
     */
    void put(const vector<uint>& pos, DataType value);




    
  protected:
 
  
  private:
    DEPOT *mDepot;    
    std::string mFileName; ;
};


template<class DataType>
UBTreeCube<DataType>::UBTreeCube(vector<uint> shape, string filename) : MultidimensionalArray(Shape),    mFileName(filename)
   {}
 
template<class DataType>
void UBTreeCube<DataType>::close() {

}
 
template<class DataType>
void UBTreeCube<DataType>::open() {

}

//Z-value UBKEY(Tuple t){
//}

template<class DataType>
uint64 UBTreeCube<DataType>::computeZIndex(const vector<uint>& idx) const throw(std::out_of_range) {
int i,s;
//the bit position in the Z-value
int bp;
Z-value addr; //the result Z-value
Bitstring bs[dimno]; // bitstring representation of the
//attribute values
//Transformation of the key attributes
for (i=0; i < d; i++) {
    // transformation of the attribute to a bitstring depends on the
    // attribute type
     bs[i] = TransformAttribute(t[i]);
}
//Bit-interleaving  Calculation of the Z-value
bp=0; //starting with the first bit of the Z-value
//looping first over dimensions then over steps realizes the
//bit-interleaving
for (s=0;s < steplength; s++) {
  for(i=0; i < d; i++) {
      // the bpth bit of the Z-value is
      // set to the sth bit of the ith bitstring
      addr[bp]=bs[i][s];
      bp++; //advance to next bit of Z-value
}}
return addr;        
        
                  //uint64 offset = 0; 
                  //for(uint i = 0; i < mNumberOfDimensions; ++i ) 
                  //  offset += idx[i] * mOffsetBase[i];
                  //if (offset >= mVolume) 
                  //  throw std::out_of_range("ubtreecube_computeoffset");
                  //return offset;
}
 
template<class DataType>
void UBTreeCube<DataType>::fillWithZeroes() {
  // I think I might go to hell for writing code like this but I could not find a way, 
  // in C to actually delete a bunch of files. 
  mChunkDescriptor.close();// not thread safe, but at least, I try to avoid kernel crashes
  ostringstream command;
  command << "rm -r -f " << mfilename<< "/[0-9]* "<< mfilename<< "/chunkdescriptor.bin";
  int ret = system(command.str().c_str());
  if(ret != 0) cerr << "Problem while removing files files in " << mfilename << endl;
  mChunkDescriptor.open();// open should make sure it recreates the file
}


template<class DataType>
UBTreeCube<DataType>::~UBTreeCube() {
  close();
} 

template<class DataType>
bool UBTreeCube<DataType>::isDense(const vector<uint>& chunkcoordinates) const {
  return (mChunkDescriptor.get(chunkcoordinates) < 0);
} 

template<class DataType>
uint UBTreeCube<DataType>::count(const vector<uint> & chunkcoordinates) const {
  if(vverbose) {
    cout << "[vverbose] count = " << abs(mChunkDescriptor.get(chunkcoordinates)) << endl;
  }
  return abs(mChunkDescriptor.get(chunkcoordinates));
}

template<class DataType>
void UBTreeCube<DataType>::incrementCount(const vector<uint> & chunkcoordinates) {
  int oldvalue = mChunkDescriptor.get(chunkcoordinates);
  if(oldvalue < 0) mChunkDescriptor.put(chunkcoordinates,oldvalue - 1);
  else {
    if(shouldItBeDense(oldvalue + 1,chunkVolume(chunkcoordinates))) {
      mChunkDescriptor.put(chunkcoordinates, - oldvalue - 1);
      makeItDense(chunkcoordinates);
    } else
      mChunkDescriptor.put(chunkcoordinates,oldvalue + 1);
  }
  if(vverbose) {
    cout << "[vverbose] increment from " << oldvalue << " to " << mChunkDescriptor.get(chunkcoordinates) << endl;
  }
 }

template<class DataType>
void UBTreeCube<DataType>::decrementCount(const vector<uint> & chunkcoordinates) {
  int oldvalue = mChunkDescriptor.get(chunkcoordinates);
  if(oldvalue < 0) { //
    if(! shouldItBeDense(oldvalue + 1, chunkVolume(chunkcoordinates) ) ) {
      mChunkDescriptor.put(chunkcoordinates,-oldvalue - 1);
      makeItSparse(chunkcoordinates);
    } else mChunkDescriptor.put(chunkcoordinates,oldvalue + 1); 
  } else mChunkDescriptor.put(chunkcoordinates,oldvalue - 1);
  if(vverbose) {
    cout << "[vverbose] decrement from " << oldvalue << " to " << mChunkDescriptor.get(chunkcoordinates) << endl;
  }
}

template<class DataType>
bool UBTreeCube<DataType>::shouldItBeDense(uint count, uint volume) {
  return count * sizeof(pair<uint,DataType>) >= volume * sizeof(DataType);
}

template<class DataType>
void UBTreeCube<DataType>::makeItDense(const vector<uint> & chunkcoordinates) {
  if(verbose) cout << "[verbose] making it dense" << endl;
  uint64 chunk_offset = mChunkDescriptor.computeOffset(chunkcoordinates);
  string path = getPath(chunk_offset);
  ifstream file( path.c_str(),std::ios::binary   ) ;
  if(!file.good()) {
        file.close();
        cerr << "Warning: file " << path << " doesn't exist, yet I'm asked to make it dense"<< endl; // great error handling
        cerr << "This is most certainly a bug as such storage would be highly inefficient." << endl;
        return;
  }
  string tmppath = path+".dense.temp";
  uint vol = chunkVolume(chunkcoordinates) * sizeof(DataType);
  FileUtil::fillFile(tmppath.c_str(), vol);
  uint actual_vol = FileUtil::getFileSize(tmppath.c_str());
  if(verbose) cout << " target vol = " << vol << " actual vol = " << actual_vol << endl;
  assert(actual_vol == vol);
  ofstream filetmp( tmppath.c_str(), std::ios::binary ); // not especially thread safe // nocreate is no longer allowed?
  uint number = count(chunkcoordinates);
  uint size_of_one_entry = sizeof(pair<uint,DataType>);
  while(number > 0) {
        uint toberead = min(number*size_of_one_entry,sizeof(mPairBuffer));
        file.read((char *) & mPairBuffer,  toberead);// no error handling
        assert(file.gcount() == (int) toberead);
        uint entries_read = toberead / size_of_one_entry;
        number -= entries_read;
        for(uint i = 0; i < entries_read; ++i) {
          filetmp.seekp(mPairBuffer[i].first * sizeof(DataType));
          filetmp.write((const char *) & mPairBuffer[i].second, sizeof(DataType));
        }
  }
  file.close();
  assert(filetmp.good());
  filetmp.close();
  if(vverbose) cout << " tmpfile size = " <<  FileUtil::getFileSize(tmppath.c_str()) << endl;
  FileUtil::copy(tmppath.c_str(),path.c_str());
  assert(FileUtil::getFileSize(path.c_str()) == vol);
  remove(tmppath.c_str());
}

template<class DataType>
void UBTreeCube<DataType>::makeItSparse(const vector<uint> & chunkcoordinates) {
  if(verbose) cout << "[verbose] making it parse" << endl;
  uint64 chunk_offset = mChunkDescriptor.computeOffset(chunkcoordinates);
  string path = getPath(chunk_offset);
  ifstream file( path.c_str(),std::ios::binary ) ;
  string tmppath = path+".sparse.temp";
  if(!file.good()) {
        file.close();
        cerr << "Warning: file " << path << " doesn't exist!"<< endl; // great error handling
        cerr << "I'm asked to make it sparse. Why wasn't it sparse before?" << endl;
        cerr << "This is probably a (harmless?) bug." << endl;
        return;
  }
  ofstream filetmp( tmppath.c_str(),std::ios::binary | std::ios::trunc ) ;
  uint buffer_size = 32;
  DataType buffer[buffer_size];
  uint actualread = 0;
  uint total_components_read = 0;
  while((actualread = file.read((char *)& buffer,buffer_size).gcount())>0) {
    uint components_read = actualread / sizeof(DataType);
    for(uint i = 0; i < components_read; ++i) {
      if(buffer[i] != 0) {
        uint component_position = total_components_read + i;
        pair<uint,DataType> entry(component_position, buffer[i]);
        filetmp.write((const char *)&entry,sizeof(entry));
      }
    }
    total_components_read += components_read;
  }
  file.close();
  assert(filetmp.good());
  filetmp.close();
  FileUtil::copy(tmppath.c_str(),path.c_str());
  remove(tmppath.c_str());
}
 
template<class DataType>
vector<uint> UBTreeCube<DataType>::getChunkedShape() const {
  return mChunkDescriptor.getShape();
} 

template<class DataType>
uint64 UBTreeCube<DataType>:: getNumberOfChunks() const {
  return mChunkDescriptor.getVolume();
}

template<class DataType>
vector<uint> UBTreeCube<DataType>::getChunkCoordinates(vector<uint> pos) const {
  for(uint i = 0; i < pos.size(); ++i) pos[i] /= mChunkShape[i];
  return pos;
}

   

template<class DataType>
DataType UBTreeCube<DataType>::get(const vector<uint>& pos) const {
  /**
   * The current code opens and close a file for each query. That's
   * probably terrible from a performance point of view so one would want
   * to pool the fstream some place.
   */
  vector<uint> chunkcoordinates = getChunkCoordinates(pos);
  bool is_dense = isDense(chunkcoordinates);
  uint64 chunk_offset = mChunkDescriptor.computeOffset(chunkcoordinates);
  uint64 offset_in_chunk =  inChunkOffset(pos,chunkcoordinates); 
  string path = getPath(chunk_offset);
  if(verbose) cout << "[verbose] Doing a get on file "<< path << endl;
  ifstream file( path.c_str(),std::ios::binary | std::ios::in) ;
  if(!file.good()) {
        file.close();
        if(verbose) 
          cout << "[verbose] File " << path << " doesn't exist yet"<< endl; // great error handling
        return DataType(0);
  }
  DataType value(0);
  if(is_dense) {
     uint local_offset = offset_in_chunk * sizeof(DataType); 
     if(verbose) cout << "[verbose] file is dense, going to position " << local_offset << endl;
     file.seekg(local_offset);
     file.read((char*) & value,sizeof(DataType));
     assert(file.gcount() == sizeof(DataType));
  } else {
      uint number = count(chunkcoordinates);
      if(verbose) cout << "[verbose] file is sparse, there are " << number << " entries." << endl;
      // in theory, we could read the pairs one by one, but in practice, this
      // would be truly bad, so I create a buffer
      uint size_of_one_entry = sizeof(pair<uint,DataType>);
      while(number > 0) {
        uint toberead = min(number*size_of_one_entry,sizeof(mPairBuffer));
        if(vverbose) cout << "[vverbose] trying to read " << toberead << " bytes. " << endl;
        file.read((char *) & mPairBuffer,  toberead);// no error handling
        if(vverbose) cout << "[vverbose]  read " << file.gcount() << " bytes. " << endl;
        assert(file.gcount() == (int) toberead);
        uint entries_read = toberead / size_of_one_entry;
        for(uint i = 0; i < entries_read; ++i) {
          if(mPairBuffer[i].first == offset_in_chunk) {
            if(verbose) cout << "[verbose] I found a match with "<< (number - i) << " left to check." <<endl;
            file.close();
            return mPairBuffer[i].second;
          }
        }
        number -= entries_read;
      }
      if(verbose) cout << "[verbose] A match was not found. I will return 0." <<endl;
  }
  assert(file.good());
  file.close();
  return value;
}


template<class DataType>
void UBTreeCube<DataType>::put(const vector<uint>& pos, DataType value) {
  // don't forget to increment counter
  vector<uint> chunkcoordinates = getChunkCoordinates(pos);
  bool is_dense = isDense(chunkcoordinates);
  uint64 chunk_offset = mChunkDescriptor.computeOffset(chunkcoordinates);
  uint64 offset_in_chunk =  inChunkOffset(pos,chunkcoordinates); 
  string path = getPath(chunk_offset);
  if(verbose) cout << "[verbose] Doing a put on file "<< path << endl;
  fstream file( path.c_str(),std::ios::binary | std::ios::out | std::ios::in  ) ;
  // if the file doesn't exist, file opening will fail
  if(!file.good()) {
    file.clear();
    file.open(path.c_str(),std::ios::binary | std::ios::out | std::ios::in | std::ios::trunc ) ;
    // should do error handling
    if(!file.good()) 
      cerr <<"[error] I can't open " << path << " for a put. " << endl;
    assert(file.good());
  }
  DataType oldvalue (0);
  if(is_dense) {// data is densely coded
     uint local_offset = offset_in_chunk * sizeof(DataType); 
     if(verbose) cout << "[verbose] file is dense, going to position " << local_offset << endl;
     file.seekg(local_offset);
     file.read((char *) & oldvalue, sizeof(DataType));// we need this for the count
     file.seekp(local_offset);
     file.write((const char*) & value,sizeof(DataType));
  } else {// data is sparsely coded
      uint number = count(chunkcoordinates);
      if(verbose) cout << "[verbose] file is sparse, there are " << number << " entries." << endl;
      // in theory, we could read the pairs one by one, but in practice, this
      // would be truly bad, so I create a buffer called mPairBuffer
      uint size_of_one_entry = sizeof(pair<uint,DataType>);
      uint64 writepos = 0;
      bool matchFound = false;
      // ugly code follows
      while((number > 0) && (! matchFound)) {
        uint toberead = min(number*size_of_one_entry,sizeof(mPairBuffer));
        file.read((char *) & mPairBuffer,  toberead);// no error handling
        if(vverbose) 
          cout << "[vverbose] file.gcount() = " << file.gcount()  << " toberead = " << toberead << endl;
        assert(file.gcount() == (int) toberead);
        uint entries_read = toberead / size_of_one_entry;
        for(uint i = 0; i < entries_read; ++i) {
          writepos += size_of_one_entry;// new line here
          if(mPairBuffer[i].first == offset_in_chunk) {
            if(verbose) cout << "[verbose] I found a match with "<< (number - i) << " left to check." <<endl;
            if(vverbose) cout << "[vverbose] i = "<< i << " number = " << number 
              << " entries_read = " << entries_read << " size_of_one_entry = " 
              << size_of_one_entry << endl;
            oldvalue = mPairBuffer[i].second;
            if( vverbose) 
              cout << file.tellg() << " - ( " << entries_read << " - " << i << " ) * " 
                << size_of_one_entry << endl;
            
            //writepos = file.tellg() -  (entries_read - i) *size_of_one_entry;
            break;
          }
        }
        number -= entries_read;
      }
      if(! matchFound) {
        //writepos = file.tellg();
        if(verbose) 
          cout << "[verbose] A match was not found. I will write the value at the end :"<< writepos <<endl;
      } else {
        if(verbose) cout << "[verbose] Match was found at position : " << writepos << endl;
      }
      file.seekp(writepos);
      pair<uint,DataType> new_entry(offset_in_chunk,value);
      file.write((char *) & new_entry, sizeof(new_entry));
  }
  assert(file.good());
  file.close();
  if(verbose) cout <<" oldvalue = " << oldvalue << " value = " << value << endl;
  if((oldvalue == 0 ) && (value != 0)) {
    if(verbose) cout << " incrementing count " << endl;
    incrementCount(chunkcoordinates);
  }
  else if ((oldvalue != 0) && (value == 0)) {    
    if(verbose) cout << " decrementing count " << endl;
    decrementCount(chunkcoordinates);      
  }
}

template<class DataType>
string UBTreeCube<DataType>::getPath(uint64 chunknumber) const {
  uint lastdigit = chunknumber % mBasis;
  uint64 whatremains = chunknumber / mBasis;
  ostringstream ost;
  ost << mfilename;
  while(whatremains > 0) {
    ost << "/" << (whatremains % mBasis);
    whatremains /= mBasis;
  }
  if(! FileUtil::fileExists(ost.str().c_str()) ) { // if directories don't exist, create them
    if( FileUtil::mkdir_parents(ost.str().c_str()) ) // make sure the path exists!!!
      cerr << " I couldn't create the proper directory. This program will fail. " 
        << endl;// great error handling  
  }
  ost << "/" << lastdigit << ".bin";
  return ost.str();
}


/**
 * Uglyness: computes the offset of the cell *inside* its chunk.
 */
template<class DataType>
uint UBTreeCube<DataType>::inChunkOffset(const vector<uint>& pos, const vector<uint> & chunkcoordinates) const {
  uint offset = 0;
  uint base = 1;
  for (uint i = 0; i < pos.size(); ++i) {
    offset += (pos[i] % mChunkShape[i] ) * base;
    base *=  min((chunkcoordinates[i]+1)*mChunkShape[i] , mShape[i] ) 
      - chunkcoordinates[i] * mChunkShape[i]; 
    // compute the "width" of the chunk
    // most time it will just b mChunkShape[i], but we have the "edges" to worry about
  }
  return offset;
}

template<class DataType>
uint UBTreeCube<DataType>::chunkVolume(const vector<uint> & chunkcoordinates) {
  uint volume = 1;
  for (uint i = 0; i < chunkcoordinates.size() ; ++i) {
    volume *= min((chunkcoordinates[i]+1)*mChunkShape[i] , mShape[i] ) 
      - chunkcoordinates[i] * mChunkShape[i]; 
  }
  return volume;
}

template<class DataType>
vector<uint> UBTreeCube<DataType>::computeChunkedShape(vector<uint> shape, const vector<uint>& chunkshape) {
  for (uint i = 0; i < shape.size(); ++i) {
    shape[i] = shape[i] / chunkshape[i] + (shape[i] % chunkshape[i] > 0 ? 1 : 0);
  }
  return shape;  
} 



/************************************************************
 * We are done with the holap cube class.
 *
 ***********************************************************/

/**
 * This is a suboptimal implementation of the iterator
 * over an Holap cube. What should be done is that the
 * iterator would run differently over dense and sparse
 * chunks and would most definitively not call the get
 * method directly on the cube. That's waaaaaayyyy too slow.
 * But for now, I'm prototyping. 
 *
 * FIX ME. MAKE ME FASTER.
 */
template <class DataType>
class CubeIterator {

  public: 

    CubeIterator(const UBTreeCube<DataType> & owner,const vector<uint>& start, const vector<uint>& end) :
      mStart(start), mEnd(end),mIndex(start),
      mCurrentValue(0),mOwner(owner){
        mIndex[0] -= 1;// start right before the area
    }

    CubeIterator(const CubeIterator<DataType> & other):
      mStart(other.mStart), mEnd(other.mEnd), mIndex(other.mIndex), 
      mCurrentValue(other.mCurrentValue), mOwner(other.mOwner)
    {}
    /**
     * This is where the code needs to get smarter.
     */
    inline bool next() {
      if(! increment ()) return false;
      while((mCurrentValue = mOwner.get(mIndex)) == 0) 
        if(! increment ()) return false;
      return true;
    }

    DataType value() const {
      return mCurrentValue;
    }

    vector<uint> key() const {
      return mIndex;
    }
    
  protected:
     inline bool increment() {
      for(uint dim = 0; dim < mIndex.size(); ++dim) {
        if(++mIndex[dim] >= mEnd[dim]) {
          mIndex[dim] = mStart[dim];
        } else {
          return true;
        }
      }
      return false;
    }   

    bool slow_increment(uint dim = 0) {
      if(++mIndex[dim] >= mEnd[dim]) {
        if(dim == mIndex.size() - 1) return false;
        mIndex[dim] = mStart[dim];
        return increment(++dim);
      }
      //cout << "increment to ";
      //for(uint k =0; k < mIndex.size() ; ++k)
      //  cout << mIndex[k] << " ";
      //cout << endl;
      return true;
    }

      
    vector<uint> mStart, mEnd, mIndex;
    DataType mCurrentValue;
    const UBTreeCube<DataType> & mOwner;
};





// #endif
// #include <depot.h>
// #include <stdlib.h>
// #include <stdio.h>
// 
// #define NAME     "mikio"
// #define NUMBER   "000-1234-5678"
// #define DBNAME   "book"
// 
// int main(int argc, char **argv){
//   DEPOT *depot;
//   char *val;
// 
//   /* open the database */
//   if(!(depot = dpopen(DBNAME, DP_OWRITER | DP_OCREAT, -1))){
//     fprintf(stderr, "dpopen: %s\n", dperrmsg(dpecode));
//     return 1;
//   }
// 
//   /* store the record */
//   if(!dpput(depot, NAME, -1, NUMBER, -1, DP_DOVER)){
//     fprintf(stderr, "dpput: %s\n", dperrmsg(dpecode));
//   }
// 
//   /* retrieve the record */
//   if(!(val = dpget(depot, NAME, -1, 0, -1, NULL))){
//     fprintf(stderr, "dpget: %s\n", dperrmsg(dpecode));
//   } else {
//     printf("Name: %s\n", NAME);
//     printf("Number: %s\n", val);
//     free(val);
//   }
// 
//   /* close the database */
//   if(!dpclose(depot)){
//     fprintf(stderr, "dpclose: %s\n", dperrmsg(dpecode));
//     return 1;
//   }
// 
//   return 0;
// }
// 
// The following example shows all records of the database.
// 
// #include <depot.h>
// #include <stdlib.h>
// #include <stdio.h>
// 
// #define DBNAME   "book"
// 
// int main(int argc, char **argv){
//   DEPOT *depot;
//   char *key, *val;
// 
//   /* open the database */
//   if(!(depot = dpopen(DBNAME, DP_OREADER, -1))){
//     fprintf(stderr, "dpopen: %s\n", dperrmsg(dpecode));
//     return 1;
//   }
// 
//   /* initialize the iterator */
//   if(!dpiterinit(depot)){
//     fprintf(stderr, "dpiterinit: %s\n", dperrmsg(dpecode));
//   }
// 
//   /* scan with the iterator */
//   while((key = dpiternext(depot, NULL)) != NULL){
//     if(!(val = dpget(depot, key, -1, 0, -1, NULL))){
//       fprintf(stderr, "dpget: %s\n", dperrmsg(dpecode));
//       free(key);
//       break;
//     }
//     printf("%s: %s\n", key, val);
//     free(val);
//     free(key);
//   }
// 
//   /* close the database */
//   if(!dpclose(depot)){
//     fprintf(stderr, "dpclose: %s\n", dperrmsg(dpecode));
//     return 1;
//   }
// 
//   return 0;
// }

#endif
